from __future__ import annotations

import abc
import inspect
import itertools
import json
import os
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, datetime
from functools import cached_property
from typing import Any, Callable, ClassVar, Generator, Iterable, Type

from google.cloud import datastore
from pydantic import BaseModel

from gcp_pilot import exceptions

DEFAULT_NAMESPACE = os.environ.get("GCP_DATASTORE_NAMESPACE", default=None)
DEFAULT_PK_FIELD_NAME = "id"
DEFAULT_PK_FIELD_TYPE = int  # pylint: disable=invalid-name
MAX_ITEMS_PER_OPERATIONS = 500  # Datastore cannot write more than 500 items per call


def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


@dataclass
class DoesNotExist(Exception):
    cls: Type[EmbeddedDocument]
    filters: dict


@dataclass
class MultipleObjectsFound(Exception):
    cls: Type[EmbeddedDocument]
    filters: dict


def _starts_with_operator(lookup_fields, value) -> list[tuple[str, str, Any]]:
    field_name = ".".join(lookup_fields)
    return [
        (field_name, ">=", value),
        (field_name, "<=", f"{value}\ufffd"),
    ]


@dataclass
class Manager:
    LOOKUP_OPERATORS: ClassVar[dict[str, (str | Callable)]] = {
        "eq": "=",
        "gt": ">",
        "gte": ">=",
        "lt": "<",
        "lte": "<=",
        "in": "in",
        "startswith": _starts_with_operator,
    }
    doc_klass: Type[EmbeddedDocument]

    @cached_property
    def client(self) -> datastore.Client:
        return datastore.Client(namespace=self.namespace)

    @property
    def namespace(self) -> str:
        return getattr(self.doc_klass.__config__, "namespace", ())

    @property
    def exclude_from_indexes(self):
        return getattr(self.doc_klass.__config__, "exclude_from_indexes", ())

    @property
    def kind(self) -> str:
        return "EmbeddedKind" if self.is_embedded else self.doc_klass.__name__

    @property
    def is_embedded(self) -> bool:
        return not issubclass(self.doc_klass, Document)

    @property
    def fields(self) -> Iterable[str]:
        return self.doc_klass.__fields__.keys()

    def build_key(self, pk: Any | None = None) -> datastore.Key:
        if self.is_embedded:
            return self.client.key(self.kind)

        if pk:
            typed_pk = DEFAULT_PK_FIELD_TYPE(pk)
            return self.client.key(self.kind, typed_pk)
        # If no primary key is provided, we let the server create a new ID
        return self.client.allocate_ids(self.client.key(self.kind), 1)[0]

    def _iterate(self, query, page_size):
        cursor = None
        empty = False

        while not empty:
            query_iter = query.fetch(start_cursor=cursor, limit=page_size)

            page = next(query_iter.pages, [])
            for item in page:
                yield item
            cursor = query_iter.next_page_token
            empty = not bool(cursor)

    def query(
        self,
        distinct_on: str | None = None,
        order_by: str | list[str] | None = None,
        page_size: int | None = None,
        **kwargs,
    ) -> datastore.query.Iterator:
        # base query
        query = self.client.query(kind=self.kind)
        if order_by:
            query.order = order_by
        if distinct_on:
            query.distinct_on = distinct_on

        # parse lookup args
        cross_filters = []
        for key, value in kwargs.items():
            all_filters = self._build_filter(key=key, value=value)
            for field_name, operator, field_value in all_filters:
                if operator == "in":
                    cross_filters.append((field_name, operator, field_value))
                else:
                    query.add_filter(field_name, operator, field_value)

        if not cross_filters:
            yield from self._iterate(query=query, page_size=page_size)
        else:
            # prepare combinations
            options = defaultdict(list)
            for name, operator, values in cross_filters:
                for value in values:
                    options[name].append((name, "=", value))

            combinations = itertools.product(*list(options.values()))

            found = set()
            for combination in combinations:
                current_query = query
                for field_name, operator, value in combination:
                    current_query = current_query.add_filter(field_name, operator, value)

                for item in self._iterate(query=current_query, page_size=page_size):
                    if item.id not in found:
                        yield item
                        found.add(item.id)

    def filter(self, **kwargs) -> Generator[Document, None, None]:
        for entity in self.query(**kwargs):
            yield self.doc_klass.from_entity(entity=entity)

    def get(self, **kwargs) -> Document:
        if DEFAULT_PK_FIELD_NAME in kwargs:
            pk = kwargs[DEFAULT_PK_FIELD_NAME]
            entity = self.client.get(key=self.build_key(pk=pk))
            if entity:
                return self.doc_klass.from_entity(entity=entity)
            raise DoesNotExist(self.doc_klass, pk)

        # Since we can't fetch directly from the key,
        # we filter and hope for just one object
        one_obj = None
        for obj in self.filter(**kwargs):
            if one_obj is not None:
                raise MultipleObjectsFound(self.doc_klass, filters=kwargs)
            one_obj = obj
        if not one_obj:
            raise DoesNotExist(self.doc_klass, filters=kwargs)
        return one_obj

    def create(self, **kwargs) -> Document:
        obj = self.doc_klass(**kwargs)
        entity = obj.to_entity()
        self.client.put(entity=entity)

        # if successfully saved, we assure the auto-generated ID is added to the final object
        if not obj.pk:
            setattr(obj, DEFAULT_PK_FIELD_NAME, entity.id)
        return obj

    def update(self, pk: str, **kwargs) -> Document:
        if kwargs:
            entity = self.client.get(key=self.build_key(pk=pk))
            # TODO: enable partial nested updates
            as_data = {
                key: value.to_dict() if isinstance(value, EmbeddedDocument) else value for key, value in kwargs.items()
            }
            entity.update(as_data)
            self.client.put(entity=entity)
        return self.get(id=pk)

    def delete(self, pk: str | None = None):
        if pk:
            self.client.delete(key=self.build_key(pk=pk))
        else:
            keys = [entity.key for entity in self.query()]
            for chunk in _chunks(keys, MAX_ITEMS_PER_OPERATIONS):
                self.client.delete_multi(keys=chunk)

    def _build_filter(self, key: str, value: Any) -> list[tuple[str, str, Any]]:
        operator = None

        *field_parts, lookup = key.split("__")
        if lookup:
            try:
                operator = self.LOOKUP_OPERATORS[lookup]

                if callable(operator):
                    field_name = ".".join(field_parts)
                    return operator(field_name, value)
            except KeyError:
                field_parts.append(lookup)

        if len(field_parts) > 1 and field_parts[0] not in self.fields:
            raise exceptions.ValidationError(
                f"{field_parts[0]} is not a valid field. Excepted one of {' | '.join(self.fields)}"
            )

        field_name = ".".join(field_parts)
        return [(field_name, (operator or "="), value)]


class EmbeddedDocument(BaseModel, abc.ABC):
    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True
        keep_untouched = (cached_property,)

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls.documents = Manager(doc_klass=cls)

    @classmethod
    def from_dict(cls, **kwargs) -> EmbeddedDocument:
        return cls(**kwargs)

    def to_dict(self) -> dict:
        return self.dict()

    @classmethod
    def from_entity(cls, entity: datastore.Entity) -> EmbeddedDocument:
        data = dict(entity.items())
        data[DEFAULT_PK_FIELD_NAME] = entity.id
        return cls.from_dict(**data)

    def to_entity(self) -> datastore.Entity:
        exclude_from_indexes = self.documents.exclude_from_indexes
        if isinstance(self, Document):
            entity = datastore.Entity(
                key=self.documents.build_key(pk=self.pk),
                exclude_from_indexes=exclude_from_indexes,
            )
            if not self.pk:
                setattr(self, DEFAULT_PK_FIELD_NAME, entity.id)
        else:
            entity = datastore.Entity(
                key=self.documents.build_key(),
                exclude_from_indexes=exclude_from_indexes,
            )

        dict_obj = json.loads(self.json())

        data = {}
        for field_name, field_info in self.__fields__.items():
            if inspect.isclass(field_info.type_) and issubclass(field_info.type_, EmbeddedDocument):
                # recursively generate entities
                field_value = getattr(self, field_name)
                if isinstance(field_value, (list, tuple)):
                    value = [item.to_entity() for item in field_value]
                else:
                    value = field_value.to_entity() if field_value else None
            else:
                value = dict_obj[field_name]  # fetch json-friendly value
            data[field_name] = value

        entity.update(data)
        return entity


class Document(EmbeddedDocument, abc.ABC):
    id: DEFAULT_PK_FIELD_TYPE | None = None

    class Config(EmbeddedDocument.Config):
        exclude_from_indexes = ()
        namespace = DEFAULT_NAMESPACE
        json_encoders = {
            datetime: lambda v: v.isoformat(),
            date: lambda v: v.isoformat(),
        }

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.update_forward_refs()  # needed when base model has optional field o.O

    @property
    def pk(self):
        return getattr(self, "id", None)

    def save(self) -> Document:
        return self.documents.create(**self.dict())

    def delete(self) -> None:
        self.documents.delete(pk=self.pk)


__all__ = (
    "DoesNotExist",
    "MultipleObjectsFound",
    "EmbeddedDocument",
    "Document",
)



# More Information: <https://cloud.google.com/logging/docs/reference/libraries#client-libraries-install-python>
from logging import INFO

from google.cloud import logging

from gcp_pilot.base import GoogleCloudPilotAPI


class CloudLogging(GoogleCloudPilotAPI):
    _client_class = logging.Client

    def _get_client_extra_kwargs(self):
        return {"project": self.project_id}

    def enable(self, log_level=INFO):
        self.client.setup_logging(log_level=log_level)

    @property
    def handler(self):
        return self.client.get_default_handler()


__all__ = ("CloudLogging",)



# More Information: https://cloud.google.com/service-usage/docs/reference/rest
from enum import Enum
from typing import Generator

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType


class ServiceStatus(Enum):
    ENABLED = "ENABLED"
    DISABLED = "Enabled"


class ServiceUsage(DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="serviceusage",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _service_path(self, service_name: str, project_id: str | None = None) -> str:
        project_path = self._project_path(project_id=project_id)
        return f"{project_path}/services/{service_name}"

    def list_services(
        self,
        project_id: str | None = None,
        status: ServiceStatus = ServiceStatus.ENABLED,
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._project_path(project_id=project_id),
        )
        if status:
            params["filter"] = f"state:{status.value}"

        yield from self._paginate(
            method=self.client.services().list,
            result_key="services",
            params=params,
        )

    def get_service(
        self,
        service_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._service_path(service_name=service_name, project_id=project_id)
        return self._execute(
            method=self.client.services().get,
            name=name,
        )

    def enable_service(
        self,
        service_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._service_path(service_name=service_name, project_id=project_id)
        return self._execute(
            method=self.client.services().enable,
            name=name,
        )

    def disable_service(
        self,
        service_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._service_path(service_name=service_name, project_id=project_id)
        return self._execute(
            method=self.client.services().disable,
            name=name,
        )



# More Information: https://cloud.google.com/api-gateway/docs/reference/rest
import base64
import time
from pathlib import Path
from typing import Generator

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI, ResourceType
from gcp_pilot.service_usage import ServiceUsage


class APIGateway(DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="apigateway",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _api_path(self, api_name: str, project_id: str | None = None) -> str:
        location_path = self._location_path(project_id=project_id, location="global")
        return f"{location_path}/apis/{api_name}"

    def _gateway_path(self, gateway_name: str, location: str | None = None, project_id: str = None) -> str:
        location_path = self._location_path(project_id=project_id, location=location)
        return f"{location_path}/gateways/{gateway_name}"

    def _config_path(self, config_name: str, api_name: str, project_id: str | None = None) -> str:
        api_path = self._api_path(api_name=api_name, project_id=project_id)
        return f"{api_path}/configs/{config_name}"

    def list_apis(
        self,
        project_id: str | None = None,
        location: str | None = "global",
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._location_path(project_id=project_id, location=location),
        )
        yield from self._list(
            method=self.client.projects().locations().apis().list,
            result_key="apis",
            params=params,
        )

    def get_api(
        self,
        api_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._api_path(api_name=api_name, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().apis().get,
            name=name,
        )

    def create_api(
        self,
        api_name: str,
        display_name: str = "",
        labels: dict[str, str] | None = None,
        project_id: str | None = None,
        wait: bool = True,
    ) -> ResourceType:

        parent = self._location_path(location="global", project_id=project_id)
        body = {
            "displayName": display_name,
            "labels": labels,
        }
        output = self._execute(
            method=self.client.projects().locations().apis().create,
            parent=parent,
            apiId=api_name,
            body=body,
        )
        if wait:
            while output.get("state", "CREATING") == "CREATING":
                output = self.get_api(api_name=api_name, project_id=project_id)
                time.sleep(5)
        return output

    def delete_api(
        self,
        api_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._api_path(api_name=api_name, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().apis().delete,
            name=name,
        )

    def list_configs(
        self,
        api_name: str,
        project_id: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._api_path(api_name=api_name, project_id=project_id),
        )
        yield from self._paginate(
            method=self.client.projects().locations().apis().configs().list,
            result_key="apiConfigs",
            params=params,
        )

    def get_config(
        self,
        config_name: str,
        api_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._config_path(config_name=config_name, api_name=api_name, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().apis().configs().get,
            name=name,
        )

    def create_config(
        self,
        config_name: str,
        api_name: str,
        service_account: str,
        open_api_file: Path,
        display_name: str = "",
        labels: dict[str, str] | None = None,
        project_id: str | None = None,
        wait: bool = True,
    ) -> ResourceType:

        parent = self._api_path(api_name=api_name, project_id=project_id)
        file_content = {"path": open_api_file.name, "contents": base64.b64encode(open_api_file.read_bytes()).decode()}
        body = {
            "displayName": display_name,
            "gatewayServiceAccount": service_account,
            "openapiDocuments": [{"document": file_content}],
            "labels": labels,
        }
        output = self._execute(
            method=self.client.projects().locations().apis().configs().create,
            parent=parent,
            apiConfigId=config_name,
            body=body,
        )

        if wait:
            while output.get("state", "CREATING") == "CREATING":
                output = self.get_config(config_name=config_name, api_name=api_name, project_id=project_id)
                time.sleep(5)
        return output

    def delete_config(
        self,
        config_name: str,
        api_name: str,
        project_id: str | None = None,
    ) -> ResourceType:
        name = self._config_path(config_name=config_name, api_name=api_name, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().apis().configs().delete,
            name=name,
        )

    def list_gateways(
        self,
        project_id: str | None = None,
        location: str | None = None,
    ) -> Generator[ResourceType, None, None]:
        params = dict(
            parent=self._location_path(project_id=project_id, location=location),
        )
        yield from self._paginate(
            method=self.client.projects().locations().gateways().list,
            result_key="gateways",
            params=params,
        )

    def get_gateway(
        self,
        gateway_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._gateway_path(gateway_name=gateway_name, location=location, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().gateways().get,
            name=name,
        )

    def create_gateway(
        self,
        gateway_name: str,
        api_name: str,
        config_name: str,
        labels: dict[str, str] | None = None,
        project_id: str | None = None,
        location: str | None = None,
        wait: bool = True,
    ) -> ResourceType:

        parent = self._location_path(project_id=project_id, location=location)
        config_path = self._config_path(
            api_name=api_name,
            config_name=config_name,
            project_id=project_id,
        )
        body = {
            "apiConfig": config_path,
            "labels": labels,
        }
        output = self._execute(
            method=self.client.projects().locations().gateways().create,
            parent=parent,
            gatewayId=gateway_name,
            body=body,
        )
        if wait:
            while output.get("state", "CREATING") == "CREATING":
                output = self.get_gateway(gateway_name=gateway_name, location=location, project_id=project_id)
                time.sleep(5)
        return output

    def delete_gateway(
        self,
        gateway_name: str,
        project_id: str | None = None,
        location: str | None = None,
    ) -> ResourceType:
        name = self._gateway_path(gateway_name=gateway_name, location=location, project_id=project_id)
        return self._execute(
            method=self.client.projects().locations().gateways().delete,
            name=name,
        )

    def enable_gateway(
        self,
        api_name: str,
        project_id: str | None = None,
    ):
        api_data = self.get_api(api_name=api_name, project_id=project_id)
        service_name = api_data["managedService"]

        service_usage = ServiceUsage()
        return service_usage.enable_service(
            service_name=service_name,
            project_id=project_id,
        )

    def disable_gateway(
        self,
        api_name: str,
        project_id: str | None = None,
    ):
        api_data = self.get_api(api_name=api_name, project_id=project_id)
        service_name = api_data["managedService"]

        service_usage = ServiceUsage()
        return service_usage.disable_service(
            service_name=service_name,
            project_id=project_id,
        )



# More Information: <https://googleapis.dev/python/secretmanager/latest/index.html>
from google.api_core.exceptions import NotFound
from google.cloud import secretmanager

from gcp_pilot.base import GoogleCloudPilotAPI


class SecretManager(GoogleCloudPilotAPI):
    _client_class = secretmanager.SecretManagerServiceClient

    def _secret_path(self, key: str, project_id: str | None = None) -> str:
        parent = self._project_path(project_id=project_id)
        return f"{parent}/secrets/{key}"

    def _secret_version_path(self, key: str, version: int | None = None, project_id: str | None = None):
        parent = self._secret_path(key=key, project_id=project_id)
        version_str = str(version) if version else "latest"
        return f"{parent}/versions/{version_str}"

    def _create_secret(self, key: str, project_id: str | None = None):
        parent = self._project_path(project_id=project_id)

        return self.client.create_secret(
            request={
                "parent": parent,
                "secret_id": key,
                "secret": {"replication": {"automatic": {}}},
            }
        )

    def _create_version(self, key: str, value: str, project_id: str | None = None):
        parent = self._secret_path(key=key, project_id=project_id)

        return self.client.add_secret_version(request={"parent": parent, "payload": {"data": value.encode()}})

    def list_secrets(
        self,
        prefix: str | None = None,
        suffix: str | None = None,
        project_id: str | None = None,
    ) -> list[tuple[str, str]]:
        parent = self._project_path(project_id=project_id)
        secrets = self.client.list_secrets(
            request={
                "parent": parent,
            }
        )
        for secret in secrets:
            name = secret.name.rsplit("secrets/", 1)[-1]
            if prefix and not name.startswith(prefix=prefix):
                continue
            if suffix and not name.endswith(suffix=suffix):
                continue
            yield name, self.get_secret(key=name, project_id=project_id)

    def add_secret(self, key: str, value: str, project_id: str | None = None) -> str:
        try:
            version = self._create_version(
                key=key,
                value=value,
                project_id=project_id,
            )
        except NotFound:
            self._create_secret(
                key=key,
                project_id=project_id,
            )
            version = self._create_version(
                key=key,
                value=value,
                project_id=project_id,
            )
        return version.name

    def get_secret(self, key: str, version: int | None = None, project_id: str | None = None) -> str:
        response = self.client.access_secret_version(
            request={"name": self._secret_version_path(key=key, version=version, project_id=project_id)}
        )
        return response.payload.data.decode()

    def rollback_secret(self, key: str, temporarily: bool = False, project_id: str | None = None):
        secret_path = self._secret_version_path(key=key, project_id=project_id)

        if temporarily:
            response = self.client.disable_secret_version(request={"name": secret_path})
        else:
            response = self.client.destroy_secret_version(request={"name": secret_path})
        return response


__all__ = ("SecretManager",)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcp_pilot', 'gcp_pilot.factories']

package_data = \
{'': ['*']}

install_requires = \
['factory_boy', 'google-api-python-client', 'pytz', 'tenacity']

extras_require = \
{'bigquery': ['google-cloud-bigquery>=3,<4'],
 'build': ['google-cloud-build>=3,<4'],
 'datastore': ['google-cloud-datastore>=2,<3', 'pydantic'],
 'dns': ['google-cloud-dns>=0,<1'],
 'healthcare': ['fhir.resources'],
 'monitoring': ['google-cloud-logging>=3,<4', 'google-cloud-error-reporting'],
 'pubsub': ['google-cloud-pubsub>=2,<3'],
 'secret': ['google-cloud-secret-manager'],
 'sheets': ['gspread>=5,<6'],
 'speech': ['google-cloud-speech>=2,<3'],
 'storage': ['google-cloud-storage>=2,<3'],
 'tasks': ['google-cloud-tasks>=2,<3', 'google-cloud-scheduler>=2,<3']}

setup_kwargs = {
    'name': 'gcp-pilot',
    'version': '1.1.0',
    'description': 'Google Cloud Platform Friendly Pilot',
    'long_description': '![Github CI](https://github.com/flamingo-run/gcp-pilot/workflows/Github%20CI/badge.svg)\n[![Coverage Status](https://coveralls.io/repos/github/flamingo-run/gcp-pilot/badge.svg)](https://coveralls.io/github/flamingo-run/gcp-pilot)\n[![python](https://img.shields.io/badge/python-3.10-blue.svg)]()\n\n# Google Cloud Pilot\n\n## Installation\n\n`pip install gcp-pilot`\n\nSome APIs need extra packages, thus you must use `extras` to add them:\n\n- Cloud Tasks: `pip install gcp-pilot[tasks]`\n- Cloud Build: `pip install gcp-pilot[build]`\n- Cloud Storage: `pip install gcp-pilot[storage]`\n- Big Query: `pip install gcp-pilot[bigquery]`\n- Speech: `pip install gcp-pilot[speech]`\n- Sheets: `pip install gcp-pilot[sheets]`\n- Pub/Sub: `pip install gcp-pilot[pubsub]`\n- Datastore: `pip install gcp-pilot[datastore]`\n- Cloud DNS: `pip install gcp-pilot[dns]`\n- Secret Manager: `pip install gcp-pilot[secret]`\n- Healthcare Engine: `pip install gcp-pilot[healthcare]`\n\n\n## Usage\n\n```\nfrom gcp_pilot.resource import ResourceManager\n\ngrm = ResourceManager()\n```\n\n## Default Values\n\n### Credentials\n\n`gcp-pilot` uses [ADC](https://cloud.google.com/docs/authentication/production#automatically) to detect credentials. This means that you must have one of the following setups:\n- Environment variable `GOOGLE_APPLICATION_CREDENTIALS` pointing to the JSON file with the credentials\n- Run inside GCP (Compute Engine, Cloud Run, GKE, AppEngine), so the machine\'s credentials will be used\n\nYou can also globally set a service account using the environment variable `DEFAULT_SERVICE_ACCOUNT`, which will require impersonation.\n\n### Project\n\nWhen creating a client, a default project is defined by using the project that the credentials belongs to.\n\nClients that support managing resources from other projects can be overwritten per call.\n\n> Example: you create a `BigQuery` client using credentials from  `project_a`.\nAll calls will query datasets from `project_a`, unless another project is passed as parameter when performing the call.\n\nYou can also globally set a project using the environment variable `DEFAULT_PROJECT`\n\n### Location\n\nVery similar to default project, a default location is defined by using the project\'s location.\nThe project\'s location will exist if you ever enabled AppEngine, so you had to set a location then.\nOtherwise, no default location will be set.\n\nYou can also globally set a location using the environment variable `DEFAULT_LOCATION` and reduce the amount of API calls \nwhen creating clients.\n\n## Why Use ``gcp-pilot``\n\n_"Since Google already has a [generic API client](https://github.com/googleapis/google-api-python-client) and so many [specific clients](https://github.com/googleapis?q=python&type=&language=), why should I use this library?"_\n\nGoogle\'s has 2 types of clients:\n- **dedicated**: custom-made for the APIs. They are excellent: they implement high level interaction with the API with friendly methods. The `gcp-pilot` can adds its value by handling authentication, friendly errors and parameter fallback.\n- **generic**: a single client that is capable of dynamically calling any REST API. They are a pain to use: very specific calls that must be translated from the documentation. The `gcp-pilot` comes in handy to add high-level interaction with friendly method such as `Calendar.create_event`, on top of all other vantages cited above.\n\n### Parameter Fallback\n\nMost API endpoints require `project_id` (sometimes even `project_number`) and `location`.\n\nSo `gcp-pilot` automatically detects these values for you, based on your credentials (although it\'ll require extra permissions and API calls).\n\nIf you use multiple projects, and your credentials is accessing other projects, you can still customize the parameters on each call to avoid the default fallback.\n\n\n### Friendly Errors\n\nMost APIs return a generic ``HttpException`` with am embedded payload with error output, and also there\'s a couple of different structures for these payloads.\n\nSo `gcp-pilot` tries its best to convert these exceptions into more friendly ones, such as `NotFound`, `AlreadyExists` and `NotAllowed`.\n\nIt\'ll be much easier to capture these exceptions and handle them by its type.\n\n\n### Identification Features\n\n- **Authentication**: each client uses [ADC](https://cloud.google.com/docs/authentication/production#automatically),\nwhich consists on trying to detect the service account with fallbacks: SDK > Environment Variable > Metadata\n- **Impersonation**: it\'s possible to create clients with ``impersonate_account`` parameter that [impersonates](https://cloud.google.com/iam/docs/impersonating-service-accounts#allow-impersonation) another account.\n- **Delegation**: services _(eg. Google Workspace)_ that requires specific subjects are automatically delegated, sometimes even performing additional credential signatures.\n- **Region**: most GCP services requires a location to work on *(some even require specific locations)*. If not provided, the clients use the project\'s default location, as defined by App Engine.\n- **Authorization**: OIDC authorization is automatically generated for services *(eg. CloudRun)* that require authentication to be used.\n\n### Auto-Authorization\n\nSome services require specific authorizations that should be setup prior to its usage, some examples:\n- [Pub/Sub] subscribe to a topic with authenticated push;\n- [Cloud Scheduler] schedule a job to trigger a Cloud Run service;\n- [Cloud Tasks] queue a task to trigger a Cloud Run service;\n\nIn these cases, `gcp-pilot` tries its best to assure that the required permissions are properly set up\nbefore the actual request is made.\n\n### Integration\n\nSome services can be integrated, and `gcp-pilot` does just that in a seamless way by adding helper methods.\n\nExample: you can subscribe to Google Cloud Build\'s events to be notified by every build step.\n\nBy using `CloudBuild.subscribe`, the `gcp-pilot` creates a subscription (and the topic, if needed) in the Google Pub/Sub service.\n\n## Supported APIs\n\n- IAM\n   - manage service accounts\n   - manage permissions\n- Identity Platform\n  - sign in users\n  - sign up users\n  - reset password flow\n  - verify email flow\n- Credentials\n  - manage API Keys\n- Resource Manager\n   - manage projects\n   - manage permissions\n- Secret Manager\n  - manage secrets\n- Identity Aware Proxy\n   - generate OIDC token\n- Source Repositories\n   - manage repositories\n- Directory:\n  - manage users\n  - manage groups\n- People:\n  - get people\n- Cloud SQL\n   - manage instances\n   - manage databases\n   - manage users\n- Cloud Storage\n   - manage buckets\n   - manage files\n- Cloud Build\n   - manage triggers\n- Cloud Functions\n  - manager functions\n  - manage permissions\n- Cloud Scheduler\n   - manage schedules\n- Cloud Tasks\n   - manage tasks & queues\n- Cloud Run\n   - read services\n   - manage domain mappings [[1]](https://cloud.google.com/run/docs/mapping-custom-domains#adding_verified_domain_owners_to_other_users_or_service_accounts)\n- API Gateway\n  - manage APIs\n  - manage API Configs\n  - manage Gateways\n- Service Usage\n  - enable/disable APIs and Services\n- BigQuery\n   - manage datasets\n   - perform queries\n- Calendar\n   - manage events\n- Google Chats\n   - build complex messages\n   - call webhook\n   - interact as bot\n- Cloud Directory\n   - manage groups\n- Cloud DNS\n   - manage DNS zones\n   - manage zone\'s registers\n- Sheets\n   - manage spreadsheets (powered by gspread)\n- Speech\n   - recognize speech from audio\n- Datastore\n   - Object Mapping ("ORM-ish" management of documents)\n- Monitoring\n  - reporting errors\n  - logging\n  - manage custom services\n- Healthcare\n  - Manage datasets\n  - Manage stores\n  - Manage FHIR resources: _powered by [fhir-resources](https://github.com/nazrulworld/fhir.resources)_\n',
    'author': 'Joao Daher',
    'author_email': 'joao@daher.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/flamingo-run/gcp-pilot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



import json
import random
from datetime import datetime, timedelta

import factory
from faker import Faker


class FirebaseUserFactory(factory.Factory):
    class Meta:
        model = dict

    uid = factory.Faker("uuid4")
    email = factory.Faker("email")
    email_verified = factory.Faker("pybool")
    display_name = factory.Faker("name")
    photo_url = factory.Faker("image_url")
    disabled = factory.Faker("pybool")
    provider_data = factory.LazyAttribute(
        lambda obj: [
            {
                "provider_id": random.choice(["google.com", "github.com"]),
                "display_name": obj.display_name,
                "photo_url": obj.photo_url,
                "email": obj.email,
                "uid": obj.uid,
            }
        ]
    )
    metadata = factory.LazyAttribute(
        lambda obj: {
            "last_sign_in_time": Faker().unix_time(),
            "creation_time": Faker().unix_time(),
        }
    )
    tenant_id = factory.Faker("word")


class FirebaseAuthTokenFactory(factory.Factory):
    class Meta:
        model = dict

    iss = factory.LazyFunction(lambda: f"https://securetoken.google.com/{Faker().word()}")
    aud = factory.Faker("url")
    iat = factory.LazyFunction(
        lambda: Faker().unix_time(
            start_datetime=datetime.utcnow(), end_datetime=datetime.utcnow() + timedelta(seconds=3600 * 24)
        )
    )
    exp = factory.LazyAttribute(lambda obj: obj.iat + 3600)
    event_id = factory.Faker("ean")
    event_type = factory.Iterator(["beforeSignUp", "beforeSignIn"])
    sign_in_method = factory.Iterator(["google.com", "github.com", "password"])
    raw_user_info = factory.LazyAttribute(
        lambda obj: json.dumps(
            {
                "name": obj.user_record["display_name"],
                "granted_scopes": " ".join([]),
                "id": obj.user_record["uid"],
                "verified_email": obj.user_record["email_verified"],
                "given_name": obj.user_record["display_name"].split(" ", 1)[0],
                "family_name": obj.user_record["display_name"].split(" ", 1)[-1],
                "locale": obj.locale,
                "hd": obj.user_record["email"].split("@")[-1],
                "email": obj.user_record["email"],
                "picture": obj.user_record["photo_url"],
            }
        )
    )
    oauth_id_token = factory.Faker("ean")
    oauth_access_token = factory.Faker("ean")
    oauth_token_secret = None
    oauth_refresh_token = factory.Faker("ean")
    oauth_expires_in = 3600
    user_agent = factory.Faker("user_agent")
    ip_address = factory.Faker("ipv4")
    locale = "en"
    sub = factory.Faker("ean")
    tenant_id = factory.Faker("word")
    user_record = factory.SubFactory(FirebaseUserFactory)
